/* Tic-Tac-Toe */

#include <stdio.h>
#include <stdlib.h>

struct Row {
	int square[3];		/* the 3 squares that make up the row */
	char *name;			/* a description of the row */
};

/*
 * Structure to rate various row configurations
 *
 * Map number of X's and O's in a row to some score
 *
 * The lower the score, the less likely you want to
 * move there.
 *
 * The higher the score, the better the move
 *
 * Config is 2 digits, representing the number of X's and O's
 * A tens digit represents the number of X's, while the
 * units digit represents the number of O's.
 * For example, a config of 11 indicates that there is 1 X,
 * 1 O, (and 1 empty square) in that row.
 */

struct Score {		/* assign score for row configs */
	int config;			/* row configuration */
   int score;			/* row rating */
};


int Board[9];			/* the actual board */
						/* 0 => empty, 1 => X, 2 => O */
char *Marker = " XO";	/* what markers go on the board */
int who=1; 				/* who is 1 for X, and 2 for O */

struct Row rows[8] = {	/* all possible 3 in a rows */
	/* first do rows */
	0, 1, 2, "top row",
	3, 4, 5, "middle row",
	6, 7, 8, "bottom row",
	/* now do columns */
	0, 3, 6, "left column",
	1, 4, 7, "middle column",
	2, 5, 8, "right column",
	/* finally, do the diagonals */
	0, 4, 8, "upper left diagonal", /* I just made these names up */
	2, 4, 6, "upper right diagonal"
};

/*
 * rate the various configurations
 *
 * the last entry contains -1, -1, indicating the end of
 * the table
 *
 * Scores are based on the assumption that we are X (1)
 * which is represented by the tens digit
 *
 * Run this game for a while and then adjust the scores
 * below until your program wins most of the games.
 * Remember, a square is part of 2 to 4 rows, so the
 * score for a square is the sum of those rows.  Sometimes
 * two or more average scores will look better than one
 * very good score.
 *
 * For example, if you can win, you always want to make
 * that move.  The way the scores are set up now, if
 * there is a square where you can get 2 2-in-a-rows,
 * and another square where you can win, the program
 * will pick the 2 2-in-a-rows move instead.  The score
 * for config 20 should be *very* high.
 *
 */
struct Score scores[] = {
	11, 01,	/* not very useful - one X and one O */
   00, 02,  /* row empty */
   01, 03,	/* one O */
   10, 04,	/* one X, this gives us 2 in a row */
   02, 05,	/* block 2 O's in a row */
   20, 06,	/* we can win with this move */
	-1, -1,	/* signal end of table */
};


/* function prototypes */
void print_board_positions(void);
void print_board_status(void);
void print_board(char *pos);
int make_move(void);
int pick_move(void);
int evaluate_row(int row);
void toggle_user(void);
int analyze_move(void);
int analyze_row(int row);

int
main()
{
	print_board_positions();
	while (1)
	{ /* loop forever */
		print_board_status();                        
		if (make_move()) break;		/* we are done */
		if (analyze_move()) break;	/* check for 3 in a row */
	}

	printf("Thanks for playing!\n");
	return (0);
}

void
print_board_positions(void)
{ /* print out the board positions */
	static char *positions = "012345678";	/* what the positions are */

	printf("Board positions are ...\n");
	/* print board with position numbers instead of X's and O's */
	print_board(positions);
}

void
print_board_status(void)
{
	char str[10];	/* string for printing board status */
	int i, n;

	/* first, build a string representing current status */
	for (i=0; i < 9; i++)
	{ /* loop through board */
		n = Board[i];			/* who is in this position */
		str[i] = Marker[n];		/* set to appropriate marker */
	}

	printf("Current board is:\n");

	/* now print this board */
	print_board(str);
}

void
print_board(char *pos)
{ /* display the board we are being passed */
	printf(" %c | %c | %c \n", pos[0], pos[1], pos[2]);
	printf("---+---+---+---\n");
	printf(" %c | %c | %c \n", pos[3], pos[4], pos[5]);
	printf("---+---+---+---\n");
	printf(" %c | %c | %c \n", pos[6], pos[7], pos[8]);
}

int
make_move(void)
{ /* get next move - pick our move or ask user for theirs */
	char buf[BUFSIZ];
	int move;

   if (who == 1) return pick_move();	/* pick our move */

   /* it is the user's turn ... ask them for their move */
	printf("Enter move for %c, or 'q' to quit, 'h' for help: ", Marker[who]);
	gets(buf);
	if (buf[0] == 'q') return 1;	/* done */
	if (buf[0] == 'h')
	{ /* help the user out */
		print_board_positions();
		return 0;
	}
	move = atoi(buf);

	/* validate user's move */
	if (move < 0 || move > 8)
	{
		printf("Move must be between 0 and 8\n");
		return 0;	/* try again */
 	}
	if (Board[move] != 0)
	{ /* square not empty */
		printf("Position %d isn't available\n", move);
		return 0;
	}

	/* move valid = make it */
	Board[move] = who;	/* indicate this move */

	toggle_user();		/* switch to other user */

	return 0;	/* normal return */
}

void
toggle_user(void)
{
	/* change who moves next */
	if (who == 1) who = 2;
	else who = 1;
}

int
analyze_move(void)
{ /* check for 3 in a row */
	int i, ret=0;	/* loop variable and return code */

	for (i=0; i < 8; i++)
	{ /* loop through all possible rows */
		printf("Checking %s ... ", rows[i].name);
		/*
		 * analyze_row will return 0, 1, or 2
		 * 0 if nobody won
		 * 1 if X won
		 * 2 if O won
		 */
		ret = analyze_row(i);	/* see if somebody won */
		if (ret != 0)
		{ /* somebody won */
			printf("%c won!\n", Marker[ret]);
		 	return ret;
		 }
		 else printf("nobody won\n");
	}

	return 0; /* nobody won */
}

int
analyze_row(int row)
{
	int numx = 0, numo = 0, i, pos, n;

	/*
	 * count the number of X's and O's in this row
	 * if we get 3 X's or 3 O's, then that person won
	 */	
	for (i=0; i < 3; i++)
	{ /* loop through all 3 cells */
		pos = rows[row].square[i];	/* what board position */
		n = Board[pos];				/* who has moved there */
		switch(n)
		{
			case 1: numx++; break;	/* X was there */
			case 2: numo++; break;	/* O was there */
		} /* we don't care about zero */
	}

	/* see if somebody won */
	if (numx == 3) return 1;		/* X won - 3 in a row */
	if (numo == 3) return 2;		/* O won - 3 in a row */
	return 0;						/* nobody won */
}

int
pick_move(void)
{ /* have program pick next move */
	int scores[9];		/* scores for each square */
   int row, score, cell, pos;
   int best_score, best_square;

   /* initialize scores table to all zeros */
   for (pos=0; pos < 9; pos++) scores[pos] = 0;

   /* loop through rows and get score for each row */
   for (row=0; row < 8; row++)
   { /* loop thru rows */
   	score = evaluate_row(row);	/* get score for row */
      /*
       * if the score is zero (indicating no good moves)
       * or negative (indicating an error)
       * then don't add this score.
       *
       * Otherwise, find each empty square in the row
       * and add the score for this row to the cummulative
       * score for that square
       */
      if (score > 0)
      { /* actual score, update score for all empty squares */
      	for (cell = 0; cell < 3; cell++)
         { /* loop through cells */
         	pos = rows[row].square[cell];
            if (Board[pos] == 0)
            { /* square is empty - add this score */
            	scores[pos] += score;
            }
         }
      }
   }

   /*
    * we have evaluated all the rows and totalled the scores
    * for all the empty squares on the board
    * pick the square with the highest score
    */
	best_score = 0;	/* no best score yet */
   best_square = -1;	/* no best square yet */
   for (pos = 0; pos < 9; pos++)
   { /* loop through squares */
   	if (scores[pos] > best_score)
      { /* better square found - save info */
      	best_score = scores[pos];
         best_square = pos;
      }
   }

   /*
    * we should now have the best square with the best score
    * if not, then there is no move
    */
    if (best_square < 0)
    { /* internal error - this shouldn't  happen */
    	printf("Can't find a square to move to!\n");
      return -1;	/* error return */
    }

	/* valid move found, make the move */
	Board[best_square] = who;	/* place marker on board */
	toggle_user();				/* switch to other user */
   return 0;					/* normal return */
}

int
evaluate_row(int row)
{ /* calculate score for given row */
	int pos, config, i;
	int numx=0, numo=0;		/* counter for X's and O's */

   /* first, determine how many X's and O's in this row */
   for (i=0; i < 3; i++)
   { /* loop thru cells in row */
   	pos = rows[row].square[i];	/* get board position */
      if (Board[pos] == 1) numx++;	/* found an X */
      else if (Board[pos] == 2) numo++; /* found an O */
   }

   /* now calculate configuration code (XO) */
   config = numx * 10 + numo;		/* 2 digit code */

   /* next, find this configuration in the table */
	for (i=0; scores[i].config >= 0; i++)
   { /* loop thru table of scores for config */
   	if (config == scores[i].config)
      { /* found entry - return value */
      	return scores[i].score;
      }
   }

   /* we didn't find any entry - it is probably full */
   if ((numx + numo) == 3) return 0;	/* no free squares */

   /* it looks like we forgot about a configuration */
   printf("Couldn't find config entry matching %d!\n", config);

   return -1;			/* error return */
}

